'use strict';

// ** Dependencies
const _ = require('lodash');
const util = require('util');
const functions = require('./functions');
const logging = require('./logging');
const errors = require('./errors');
const Promise = require('bluebird');
const EventEmitter = require('events').EventEmitter;

class Task extends EventEmitter {
    constructor(name, handler, options) {
        super();

        this.name = name;
        this.handler = handler;
        this.options = options;
        this.parameters = options.parameters || {};
        this.logger = logging.createLogger(name);
    }

    run(args) {

        const logger = this.logger;
        const func = this.handler;
        const parameters = this.parameters;

        // Build list of arguments to pass to the handler
        const parameterNames = functions.getParameterNames(func);
        logger.debug('PARAMETERS', parameterNames);

        // Return a function that will call this task with the specified args
        return new Promise((resolve, reject) => {

            // Argument defaults
            args = args || {};
            logger.debug('ARGS', args);

            // Build list of arguments to pass to the command
            const argList = _.map(parameterNames, param => {
                const parameter = parameters[param]; // Parameter definition
                const value = args[param];

                // Validate parameter
                if (parameter) {
                    logger.debug('VALIDATE_PARAMETER', parameter);

                    // Required Parameter Support
                    if (parameter.required) {
                        if (_.isEmpty(value)) {
                            const defaultValue = parameter["default"];

                            // Apply Default Value
                            if (defaultValue) {
                                logger.debug('APPLY_DEFAULT', {parameter: param, value: defaultValue});
                                return defaultValue;
                            }

                            throw errors('ARGUMENT_REQUIRED', `The argument "${param}" is required.`, {
                                name: param,
                                value: value
                            });
                        }
                    }

                    // Type Validation
                    if (parameter.type) {
                        switch (parameter.type.toLowerCase()) {
                            case "string":
                                if (!_.isString(value)) {
                                    if (!parameter.convert)
                                        throw errors('ARGUMENT_TYPE_ERROR', 'The argument must be a valid string', {
                                            parameter: param,
                                            value: value
                                        });

                                    // Convert value to string
                                    logger.debug('Converting parameter value to string', {
                                        parameter: param,
                                        value: value
                                    });
                                    return value.toString();
                                }
                                break;
                            default:
                                throw errors(
                                    'INVALID_PARAMETER_TYPE',
                                    'The parameter type is not supported', {type: parameter.type})
                        }
                    }
                }

                return value;
            });

            try {
                // Call the target with the arguments
                const result = func.apply(func, argList);
                resolve(result);
            } catch (err) {
                reject(err);
            }
        });
    }
}

function isTask(value) {
    return value instanceof Task;
}

/**
 * Create a new task
 * @param name
 * @param handler
 * @param options
 * @returns {*}
 */
function createTask(name, handler, options) {

    // Wrap a function directly
    if (arguments.length === 1 && util.isFunction(arguments[0]))
        return createTask(arguments[0].name, arguments[0]);

    // All tasks must have a name
    if (!util.isString(name))
        throw errors('ARGUMENT_ERROR', 'The task name must be a string', {name: name});

    // Ensure func is a function
    if (!util.isFunction(handler))
        throw errors('ARGUMENT_ERROR', 'The task handler must be a function.', {func: handler});

    // Argument defaults
    options = options || {};

    // Create a new task
    const task = new Task(name, handler, options);
    return task;
}

// ** Exports
module.exports = createTask;
module.exports.isTask = isTask;
module.exports.Task = Task;